#include "cqthreadhelper.h"
#include <pthread.h>

CQThreadHelper::CQThreadHelper(QObject *parent) :
    QThread(parent)
{
    // Thread 동작하지 않음
    m_bRunning      = false;

    // 본 class의 instance를 가지는 thread를 저장
    // 본 class instance 하나를 다른 thread끼리 공유하는 것을 방지하기 위함임.
    // (최대한 동기화 관련 이슈를 원천 봉쇄)
    m_nCallingTID   = pthread_self();

    // 생성된 thread의 tid
    m_nThreadTID    = 0;

    // 실행 변수 초기화
    m_pfnThreadProc = NULL;
    m_pContext      = NULL;
}

void CQThreadHelper::run()
{
    // 시작됨
    m_bRunning = true;

    if (NULL == m_pfnThreadProc)
    {
        ERRORLOG("Invalid Thread Proc");
        goto FINAL;
    }

    // 생성된 thread의 tid를 기록한다.
    m_nThreadTID = pthread_self();

    INFOLOG("Thread(%lu) created", (unsigned long int)m_nThreadTID);
    // Thread Procedure를 호출한다.
    (*m_pfnThreadProc)(m_pContext);

FINAL:
    // 종료됨
    m_bRunning = false;
    INFOLOG("Thread(%lu) finished", (unsigned long int)m_nThreadTID);
}

bool CQThreadHelper::BeginThread(IN PFN_THREAD_PROC pfnThreadProc, IN void* pContext)
{
    bool bRtnValue = false;

    if ((NULL == pfnThreadProc) || (NULL == pContext))
    {
        bRtnValue = false;
        goto FINAL;
    }

    if (0 == pthread_equal(m_nCallingTID, pthread_self()))
    {
        // instance를 생성한 caller thread와 다른 thread에서 호출을 방지한다.
        bRtnValue = false;
        ERRORLOG("Invalid caller thread, %lu, %lu", (unsigned long int)m_nCallingTID, (unsigned long int)pthread_self());
        goto FINAL;
    }

    if (true == m_bRunning)
    {
        bRtnValue = false;
        ERRORLOG("Already Running");
        goto FINAL;
    }

    // 실행 변수 할당
    m_pfnThreadProc = pfnThreadProc;
    m_pContext      = pContext;

    INFOLOG("Try to start thread, parent=%lu", (unsigned long int)m_nCallingTID);

    // 쓰레드를 시작한다.
    // CQThreadHelper::run()이 호출된다.
    // 그리고 아래 함수는 block 함수가 아니다.
    start();

    // 여기까지 왔다면 성공
    bRtnValue = true;

FINAL:
    return bRtnValue;
}

void CQThreadHelper::WaitThread(void)
{
    if (0 == pthread_equal(m_nCallingTID, pthread_self()))
    {
        // instance를 생성한 caller thread와 다른 thread에서 호출을 방지한다.
        ERRORLOG("Invalid caller thread, %lu, %lu", (unsigned long int)m_nCallingTID, (unsigned long int)pthread_self());
        goto FINAL;
    }

    // 종료시까지 대기한다.
    INFOLOG("Wait for thread(%lu)", (unsigned long int)m_nThreadTID);
    wait(ULONG_MAX);
    INFOLOG("Finished to wait for thread(%lu)", (unsigned long int)m_nThreadTID);

FINAL:
    return;
}

// UI에서 Thread를 생성할때, UI쪽에서 signal과 slot을 정의하고 bind한다.
// 조건은,
//      1) BeginThread의 Thread Proc이 UI(QWidget*)의 static function 이다.
//      2) Signal과 Slot이 1)의 QWidget에 정의되어 있다.
//      3) Thread의 context로 UI(QWidget*) 정보가 전달되어야 한다.
// 이다.
//
// 사용예는 다음과 같다. 주요 함수의 리턴 체크는 Skip함.
/*

< .h >
class Qxxxxx : QWidget
{
...
static void THREAD_PROC(void* p);
...
public slots:
    void OnMyMsg();

signals:
    void WM_SIGNAL_MY_MSG();
...
private:
    CQThreadHelper m_cThread;
...
}

< .cpp >

void Qxxxxx::somefunction()
{
    // Message Handler(slot) 등록
    ON_MESSAGE_POST(m_cThread, SIGNAL(WM_SIGNAL_MY_MSG()), SLOT(OnMyMsg()));

    // Start Thread
    m_cThread.BeginThread(Qxxxxx::THREAD_PROC, this);
}

// Thread Procedure
void Qxxxxx::THREAD_PROC(void* p)
{
    ...
    // Send Message
	// emit하면 Primary UI Thread로 호출되지 않을 수 있다.
    // emit ((Qxxxxx*)p)->WM_SIGNAL_MY_MSG();
	// Parameter가 있는 경우는 아래와 같다.
	// QMetaObject::invokeMethod((Qxxxxx*)p, "WM_MESSAGE", Qt::BlockingQueuedConnection, Q_ARG(int, i), Q_ARG(int, i+1), ...);
	QMetaObject::invokeMethod((Qxxxxx*)p, "WM_MESSAGE", Qt::BlockingQueuedConnection);
    ...
}

// Slot (Message Handler)
void Qxxxxx::OnMyMsg()
{
    // do something,...
}
*/
